﻿// CodeMirror, copyright (c) by Marijn Haverbeke and others
// Distributed under an MIT license: http://codemirror.net/LICENSE

(function (mod) {
    if (typeof exports == "object" && typeof module == "object") // CommonJS
        mod(require("codemirror/lib/codemirror"));
    else if (typeof define == "function" && define.amd) // AMD
        define(["../../lib/codemirror"], mod);
    else // Plain browser env
        mod(CodeMirror);
})(function (CodeMirror) {
    var Pos = CodeMirror.Pos;

    function forEach(arr, f) {
        for (var i = 0, e = arr.length; i < e; ++i) f(arr[i]);
    }

    function arrayContains(arr, item) {
        if (!Array.prototype.indexOf) {
            var i = arr.length;
            while (i--) {
                if (arr[i] === item) {
                    return true;
                }
            }
            return false;
        }
        return arr.indexOf(item) != -1;
    }

    function scriptHint(editor, keywords, getToken, options) {

        // Find the token at the cursor
        var cur = editor.getCursor(), token = getToken(editor, cur);
        if (/\b(?:string|comment)\b/.test(token.type)) return;
        token.state = CodeMirror.innerMode(editor.getMode(), token.state).state;

        // If it's not a 'word-style' token, ignore the token.
        if (!/^[\w$_]*$/.test(token.string)) {
            token = {
                start: cur.ch, end: cur.ch, string: "", state: token.state,
                type: token.string == "." ? "property" : null
            };
        } else if (token.end > cur.ch) {
            token.end = cur.ch;
            token.string = token.string.slice(0, cur.ch - token.start);
        }

        var tprop = token;
        // If it is a property, find out what it is a property of.
        while (tprop.type == "property") {
            tprop = getToken(editor, Pos(cur.line, tprop.start));
            if (tprop.string != ".") return;
            tprop = getToken(editor, Pos(cur.line, tprop.start));
            if (!context) var context = [];
            context.push(tprop);
        }
        return {
            list: getCompletions(token, context, keywords, options),
            from: Pos(cur.line, token.start),
            to: Pos(cur.line, token.end)
        };
    }

    function javascriptHint(editor, options) {
        return scriptHint(editor, javascriptKeywords,
            function (e, cur) { return e.getTokenAt(cur); },
            options);
    };
    CodeMirror.registerHelper("hint", "javascript", javascriptHint);

    function getCoffeeScriptToken(editor, cur) {
        // This getToken, it is for coffeescript, imitates the behavior of
        // getTokenAt method in javascript.js, that is, returning "property"
        // type and treat "." as indepenent token.
        var token = editor.getTokenAt(cur);
        if (cur.ch == token.start + 1 && token.string.charAt(0) == '.') {
            token.end = token.start;
            token.string = '.';
            token.type = "property";
        }
        else if (/^\.[\w$_]*$/.test(token.string)) {
            token.type = "property";
            token.start++;
            token.string = token.string.replace(/\./, '');
        }
        return token;
    }

    function coffeescriptHint(editor, options) {
        return scriptHint(editor, coffeescriptKeywords, getCoffeeScriptToken, options);
    }
    CodeMirror.registerHelper("hint", "coffeescript", coffeescriptHint);

    var stringProps = ("charAt charCodeAt indexOf lastIndexOf substring substr slice trim trimLeft trimRight " +
        "toUpperCase toLowerCase split concat match replace search").split(" ");
    var arrayProps = ("length concat join splice push pop shift unshift slice reverse sort indexOf " +
        "lastIndexOf every some filter forEach map reduce reduceRight ").split(" ");
    var funcProps = "prototype apply call bind".split(" ");
    var javascriptKeywords = ("break case catch continue debugger default delete do else false finally for function " +
        "if in instanceof new null return switch throw true try typeof var void while with").split(" ");
    var coffeescriptKeywords = ("and break catch class continue delete do else extends false finally for " +
        "if in instanceof isnt new no not null of off on or return switch then throw true try typeof until void while with yes").split(" ");

    var codeMirrorCurrentColumns;
    var lastSchemaId;
    var lastVer = 0;

    function setupColumns(options) {
        if (options.schema && (!codeMirrorCurrentColumns || lastSchemaId !== options.schema.id || lastVer !== options.schema.ver)) {
            lastSchemaId = options.schema.id;
            lastVer = options.schema.ver;
            codeMirrorCurrentColumns = [];
            if (options.schema.items) {
                for (var i = 0; i < options.schema.items.length; i++) {
                    codeMirrorCurrentColumns.push('<[' + options.schema.items[i].header + ' (' + options.schema.items[i].name + ')]>');
                }
            }
        }
        return codeMirrorCurrentColumns;
    };

    var codeMirrorCurrentUserColumns;
    function setupUserColumns(options) {
        if (!codeMirrorCurrentUserColumns) {
            codeMirrorCurrentUserColumns = [];
            if (options.currentUser) {
                for (var i = 0; i < options.currentUser.items.length; i++) {
                    codeMirrorCurrentUserColumns.push('<[' + options.currentUser.items[i].header + ' (' + options.currentUser.items[i].name + ')]>');
                }
            }
            codeMirrorCurrentUserColumns.push(['login','roles','email','currentLanguageCode']);
        }
        return codeMirrorCurrentUserColumns;
    };

    function getCompletions(token, context, keywords, options) {
        var found = [], start = token.string, global = options && options.globalScope || window;
        function maybeAdd(str) {
            if (str.lastIndexOf(start, 0) == 0 && !arrayContains(found, str)) found.push(str);
        }
        function gatherCompletions(obj) {
            if (typeof obj == "string") forEach(stringProps, maybeAdd);
            else if (obj instanceof Array) forEach(arrayProps, maybeAdd);
            else if (obj instanceof Function) forEach(funcProps, maybeAdd);
            for (var name in obj) maybeAdd(name);
        }

        if (context && context.length) {
            // If this is a property, see if it belongs to some object we can
            // find in the current environment.
            var obj = context.pop(), base;
            if ((obj.type && obj.type.indexOf("columnname") === 0) ||
                (obj.string === 'doo' && context && context.length === 1 
                && context[0].string === 'changedItem' && context[0].type === 'property')) {
                var ourKeywordsCFunc = ("value setValue(value) originalValue isVisible isEnabled isRequired currentlyChanged filterForDropdown label").split(" ");
                forEach(ourKeywordsCFunc, maybeAdd);
            }
            else if (obj.type && obj.type.indexOf("variable") === 0) {
                if (options && options.additionalContext)
                    base = options.additionalContext[obj.string];
                if (!options || options.useGlobalScope !== false)
                    base = base || global[obj.string];
                if (obj.string === 'doo' && context && context.length === 1
                    && context[0].string === 'model' && context[0].type === 'property') {
                    var ourKeywordsC = ("isValid isReadOnly infotext hideCategory('categoryName') showCategory('categoryName')").split(" ");
                    forEach(ourKeywordsC, maybeAdd);
                    ourKeywordsC = setupColumns(options);
                    forEach(ourKeywordsC, maybeAdd);
                }
                if (obj.string === 'doo' && context && context.length === 1
                    && context[0].string === 'alert' && context[0].type === 'property') {
                    var ourKeywordsAl = ("showInfo('Title','Text') showWarning('Title','Text') showError('Title','Text') showSuccess('Title','Text') showQuestion('Title','Text')").split(" ");
                    forEach(ourKeywordsAl, maybeAdd);
                }
                if (obj.string === 'doo' && context && context.length === 1
                    && context[0].string === 'toast' && context[0].type === 'property') {
                    var ourKeywordsAl = ("info('Text','Title') warning('Text','Title') error('Text','Title') success('Text','Title')").split(" ");
                    forEach(ourKeywordsAl, maybeAdd);
                }
                if (obj.string === 'doo' && context && context.length === 1
                    && context[0].string === 'functions' && context[0].type === 'property') {
                    var ourKeywordsFnc = ("math date request").split(" ");
                    forEach(ourKeywordsFnc, maybeAdd);
                }
                if (obj.string === 'doo' && context && context.length === 1
                    && context[0].string === 'table' && context[0].type === 'property') {
                    var ourKeywordsFncM = ("getData(tableNameOrId,options,applicationId?) getCount(tableNameOrId,options,applicationId?) getParameter(tableNameOrId,fieldName,applicationId?) getRecord(tableNameOrId,recordId,applicationId?) getLinkedRecordsV2(tableNameOrId,recordId,fieldName,options?) getLinkedRecords(tableNameOrId,recordId,linkTableNameOrId,linkFieldName,applicationId?) createRecord(tableNameOrId,fields,options?) updateFields(tableNameOrId,recordId,fields,options?) updateRecord(tableNameOrId,recordId,fields,options?) deleteRecord(tableNameOrId,recordId,options?) createRecordsBulk(tableNameOrId,fieldsArray,options?) updateRecordsBulk(tableNameOrId,fieldsArray,options?) deleteRecordsBulk(tableNameOrId,limit,filter,options?) getFile(tableNameOrId,fileId,applicationId?) getThumbnail(tableNameOrId,fileId,applicationId?) reloadUserData(tableNameOrId)").split(" ");
                    forEach(ourKeywordsFncM, maybeAdd);
                }                
                if (obj.string === 'doo' && context && context.length === 1
                    && context[0].string === 'environment' && context[0].type === 'property') {
                    var ourKeywordsFnc = ("isPublicForm queryParam currentApplication").split(" ");
                    forEach(ourKeywordsFnc, maybeAdd);
                }
                if (obj.string === 'doo' && context && context.length === 1
                    && context[0].string === 'workflow' && context[0].type === 'property') {
                    var ourKeywordsFnc = ("stop()").split(" ");
                    forEach(ourKeywordsFnc, maybeAdd);
                }
                if (obj.string === 'doo' && context && context.length === 2
                    && context[0].string === 'currentApplication' && context[0].type === 'property') {
                        var ourKeywordsFncM = ("id header").split(" ");
                    forEach(ourKeywordsFncM, maybeAdd);
                }
                if (obj.string === 'doo' && context && context.length === 2
                    && context[0].string === 'math' && context[0].type === 'property') {
                    var ourKeywordsFncM = ("round(123.456,1)").split(" ");
                    forEach(ourKeywordsFncM, maybeAdd);
                }
                if (obj.string === 'doo' && context && context.length === 2
                    && context[0].string === 'date' && context[0].type === 'property') {
                    var ourKeywordsFncM = ("addDays(someDate,7)").split(" ");
                    forEach(ourKeywordsFncM, maybeAdd);
                }
                if (obj.string === 'doo' && context && context.length === 2
                    && context[0].string === 'request' && context[0].type === 'property') {
                    var ourKeywordsFncM = ("get('http[s]://url',{headers:,params:,basicAuth:,bearer:,...}) post('http[s]://url',{mybodydata:'abc'},{headers:,params:,basicAuth:,bearer:,...}) custom({method:'put/...',url:'http[s]://url',body:,headers:,params:,basicAuth:,bearer:,...})").split(" ");
                    forEach(ourKeywordsFncM, maybeAdd);
                }
                if (obj.string === 'doo' && context && context.length === 1
                    && context[0].string === 'form' && context[0].type === 'property') {
                    var ourKeywordsFncM = ("openForm(tableNameOrId,options,applicationId)").split(" ");
                    forEach(ourKeywordsFncM, maybeAdd);
                }
                if (obj.string === 'doo' && context && context.length === 1
                    && context[0].string === 'currentUser' && context[0].type === 'property') {
                    var ourKeywordsCu = setupUserColumns(options);
                    forEach(ourKeywordsCu, maybeAdd);
                }
                if (obj.string === 'doo' && context.length === 0) {
                    if (options && options.isWorkflow) {
                        var ourKeywordsDoo = ("model alert toast currentUser changedItem functions table form environment workflow").split(" ");
                    } else {
                        var ourKeywordsDoo = ("model alert toast currentUser changedItem functions table form environment").split(" ");
                    }
                    forEach(ourKeywordsDoo, maybeAdd);
                }
            } else if (obj.type == "string") {
                base = "";
            } else if (obj.type == "atom") {
                base = 1;
            } else if (obj.type == "function") {
                if (global.jQuery != null && (obj.string == '$' || obj.string == 'jQuery') &&
                    (typeof global.jQuery == 'function'))
                    base = global.jQuery();
                else if (global._ != null && (obj.string == '_') && (typeof global._ == 'function'))
                    base = global._();
            }
            while (base != null && context.length)
                base = base[context.pop().string];
            if (base != null) gatherCompletions(base);
        } else {
            // If not, just look in the global object and any local scope
            // (reading into JS mode internals to get at the local and global variables)
            if (options && options.isWorkflow) {
                var ourKeywords = ("doo doo.model doo.alert doo.toast doo.currentUser doo.changedItem doo.functions doo.table doo.environment doo.workflow --------------------").split(" ");
            } else {
                var ourKeywords = ("doo doo.model doo.alert doo.toast doo.currentUser doo.changedItem doo.functions doo.table doo.environment --------------------").split(" ");
            }
            forEach(ourKeywords, maybeAdd);

            // for (var v = token.state.localVars; v; v = v.next) maybeAdd(v.name);
            // for (var v = token.state.globalVars; v; v = v.next) maybeAdd(v.name);
            // if (!options || options.useGlobalScope !== false)
            //    gatherCompletions(global);
            forEach(keywords, maybeAdd);
        }
        return found;
    }
});
